# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


import numpy
from numba import prange, njit, uint64, uint32, int64, uint16, uint8


@njit(inline="always")
def binomial_kernel(k, dtype):
    """
    Creates a 1-dimensional binomial kernel.
    The coefficients in this kernel approximate a Gaussian distribution.

    :param k: The length of the kernel.

    :param dtype: The :py:class:`numpy:numpy.dtype` of the kernel array.

    :return: A :py:class:`numpy:numpy.ndarray` of shape :code:`(k, )` and integer type.
    """
    if k < 0:
        raise ValueError("The given kernel size must not be negative!")
    kernel = numpy.zeros(k, dtype=dtype)
    kernel[0] = 1
    for i in range(k - 1):
        w = kernel[0]
        for j in range(k - 1):
            w_new = kernel[1 + j]
            kernel[1 + j] += w
            w = w_new
    return kernel


@njit([(uint64, uint8[:, :, ::1], uint8[:, :, ::1], uint8[:, :, ::1]),
       (uint64, uint16[:, :, ::1], uint16[:, :, ::1], uint16[:, :, ::1])],
      parallel=True, fastmath=True, nogil=True, cache=True)
def __filter_2d_nbinomial(k, signal, buffer, out):
    """
    Convolves a batch of 2D signals with a two-dimensional normalized binomial kernel.
    The signals are padded by zeros.

    :param k: The size of the binomial kernel to use.

    :param signal: An integer array of shape :code:`(n, h, w)`. It represents the batch of 2D signals to be filtered.

    :param buffer: An integer array of shape :code:`(n, h, w)`. This procedure may write undefined values to it.

    :return: The filtered version of the array.
    """

    h, w, d = signal.shape
    if h * w * d == 0:
        return

    # Note: This procedure is not particularly optimized, because k is so general.
    kernel = binomial_kernel(k, uint32)

    # Kernel horizontal, output horizontal:
    for y in prange(h):
        for x in prange(w):
            for c in range(d):
                r = uint32(0)
                for kidx in range(k):
                    index = int64(x) - int64(k // 2) + kidx
                    index_padded = max(0, min(w - 1, index))
                    r += kernel[kidx] * (index_padded == index) * signal[y, index_padded, c]
                buffer[y, x, c] = r // 2 ** (k - 1)

    # Kernel vertical, output horizontal:
    for y in prange(h):
        for x in prange(w):
            for c in range(d):
                r = uint32(0)
                for kidx in range(k):
                    index = int64(y) - int64(k // 2) + kidx
                    index_padded = max(0, min(h - 1, index))
                    r += kernel[kidx] * (index_padded == index) * buffer[index_padded, x, c]
                out[y, x, c] = r // 2 ** (k - 1)


def filter_2d_nbinomial(k, signal):
    """
    Convolves a batch of 2D signals with a two-dimensional normalized binomial kernel.
    The signals are padded by zeros.

    :param k: The size of the binomial kernel to use.

    :param signal: An integer array of shape :code:`(n, h, w)`. It represents the batch of 2D signals to be filtered.

    :return: The filtered version of the array.
    """

    if signal.size == 0:
        return signal.copy()
    buffer = numpy.empty_like(signal)
    out = numpy.empty_like(signal)
    __filter_2d_nbinomial(uint64(k), signal, buffer, out)
    return out




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


import zipfile

from apfio.base.types import check_type
from apfio.base.zip import ZippedArray
from apfio.imaging.avif import AVIFArray
from apfio.imaging.pil import PILArray
from apfio.vision.format import ImageFormat


def vopen(file, mode='r', shape=None, fmt=None, max_readers=1, zipargs=None, array_type=None, **eparams):
    r"""
    Opens a \*.zip file for reading and/or writing compressed image files from it.

    :param file: A path or file-like object representing the \*.zip file to open, as accepted by :py:class:`~python:zipfile.ZipFile`.

    :param mode: The mode in which the file is to be opened, see :py:class:`~python:zipfile.ZipFile`.

    :param shape: The shape of the array to open.

    :param fmt: An :py:class:`~apfio.vision.format.ImageFormat` supported by :py:mod:`apfio.vision`.

    :param max_readers: The maximum number of zip file readers to use for concurrent reading from the zip file.

    :param zipargs: Either `None`, or a :py:class:`~python:dict` of additional keyword arguments to be passed on to the :py:class:`~python:zipfile.ZipFile`
                    constructor.

    :param array_type: Either `None`, or a type that is a subclass of :py:class:`~apfio.imaging.array.CompressedImageArray`. If given, this procedure
                       will use this type to convert images to bitstrings and vice versa. If omitted, this procedure
                       will automatically select a :py:class:`~apfio.imaging.array.CompressedImageArray` subclass.

    :param eparams: Keyword arguments to be passed on to the image encoder (if images are to be written). For some
                    formats this will be :py:meth:`PIL.Image.save`, for others it might be FFMPEG command line arguments.

    :return: An :py:class:`~apfio.base.array.IndependentPointArray` object the data points of which are three-dimensional numpy arrays. The object
             can safely be accessed concurrently and it is a context manager.
    """

    if mode == 'a':
        raise ValueError("File mode 'a' is not supported, as this mode would require modifying existing entries in"
                         "the zip file, which cannot be done efficiently.")

    if mode not in 'rwx':
        raise ValueError("'mode' can only be 'r', 'w', or 'x'!")

    fmt = check_type(fmt, ImageFormat, allow_none=True)

    if zipargs is None:
        zipargs = dict()

    zf = None
    zarray = None
    iarray = None

    try:
        zf = zipfile.ZipFile(file, mode=mode, **zipargs)
        zarray = ZippedArray(zf, read=mode == 'r', write=mode in 'wx', own=True, max_open=max_readers)
        if fmt is None:
            fmt = ImageFormat.from_ext(zarray.extension)
        else:
            if zarray.write:
                zarray.extension = fmt.extension
            else:
                if zarray.extension != fmt.extension:
                    raise ValueError(f"The 'fmt' parameter for 'vopen' was set to {fmt.extension}, but the zip file "
                                     f"actually uses '{zarray.extension}' as the filename extension for its entries!")

        if shape is None:
            shape = zarray.shape
        else:
            if zarray.write:
                zarray.shape = shape
            else:
                if zarray.shape != shape:
                    raise ValueError(f"The 'shape' parameter for 'vopen' was set to {shape}, but the zip file "
                                     f"actually indicates '{zarray.shape}'!")

        if array_type is None:
            if fmt == ImageFormat.AVIF:
                iarray = AVIFArray(zarray, own=True)
            else:
                iarray = PILArray(zarray, own=True)
        else:
            iarray = array_type(zarray, own=True)
        assert iarray.shape == shape
        if iarray.writable:
            iarray.format = fmt.identifier
        else:
            if iarray.format != fmt.identifier:
                raise IOError(f"The *.zip file indicates '{zarray.extension}' as the filename extension for its entries, "
                              f"but at the same time it indicates '{iarray.format}' as the image format, which does "
                              f"not seem to fit together!")
        iarray.encoder_config = eparams

        return iarray

    except:
        if iarray is not None:
            iarray.close()
        if zarray is not None:
            zarray.close()
        if zf is not None:
            zf.close()
        raise



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


import numpy

from apfio.vision.binomial import filter_2d_nbinomial

def filter_dark(image, k=16, t=10, d0=5, d1=20):
    """
    Smoothes and tone-maps dark regions of an image.

    :param image: An integer :py:class:`numpy:numpy.ndarray` of shape :code:`(h, w, d)` that is to be filtered.

    :param k: The size of the smoothing kernel to use. :code:`k = 1` disables smoothing. Larger `k` means stronger smoothing.

    :param t: The pixel brightness (average of all color channels) under which smoothing is to occur. Pixels brighter
              than this will not be modified by smoothing. The darker a pixel, the stronger the effect of smoothing.

    :param d0: Pixels with a brightness lower than this after smoothing this will be set to zero.

    :param d1: Pixels with a brightness lower than this but higher than d0 after smoothing will be adjusted by a linear
               transformation such that the darkest pixels with be set to zero and the brightest pixels will be left
               unmodified. Pixels brighter than d1 will not be modified at all.

    :return: The filtered image.
    """

    dtype = image.dtype

    if dtype == numpy.uint8:
        num_bits = 8
        double = numpy.uint16
    elif dtype == numpy.uint16:
        num_bits = 16
        double = numpy.uint32
    else:
        raise NotImplementedError(f"filter_dark cannot handle dtype {dtype}!")

    m = 2 ** num_bits - 1

    h, w, d = image.shape

    smooth = filter_2d_nbinomial(k, image)
    smooth = smooth.astype(double)

    alpha = smooth.sum(axis=-1, keepdims=True) // d
    alpha = numpy.where(alpha < t, (m * alpha) // t, m)

    interpol = alpha * image.astype(double) + (m - alpha) * smooth
    interpol = (interpol // m).astype(dtype)

    toned = numpy.where(interpol < d1, ((d1 - 0) * (interpol.clip(d0) - d0)) // (d1 - d0), interpol)

    return toned




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


from enum import Enum


class ImageFormat(Enum):
    """
    Enumerates image file formats supported by :py:mod:`apfio.vision`.
    """
    AVIF = 0 # Best AVIF implementation available.
    JPEG = 1 # Best JPEG implementation available.
    PNG = 2 # Best PNG implementation available.
    _ignore_ = ('__ext', '__identifier')
    __ext = [".avif", ".jpg", ".png"]
    __identifier = ['AVIF', 'JPEG', 'PNG']

    @property
    def extension(self):
        """
        The file name extension used for this format.

        :return: A string starting with a dot.
        """
        return ImageFormat.__ext[self.value]

    @property
    def identifier(self):
        """
        The identifier string for serializing this format.

        :return: A str.
        """
        return ImageFormat.__identifier[self.value]

    @staticmethod
    def from_ext(ext):
        """
        Turns a file name extension into an :py:class:`~apfio.vision.format.ImageFormat` object.

        :param ext: The filename extension to find the :py:class:`~apfio.vision.format.ImageFormat` for. Should include a leading dot.

        :raises ValueError: If the given filename extension was not assigned a definitive :py:class:`~apfio.vision.format.ImageFormat`.

        :return: An :py:class:`~apfio.vision.format.ImageFormat` object.
        """
        try:
            return ImageFormat(ImageFormat.__ext.index(ext))
        except IndexError:
            raise ValueError(f"apfio.vision does not support any image format for the filename extension '{ext}'!")